
#include <triangulation/dim3.h>
#include <cstring>
#include <map>
#include <queue>
// #include <pthread.h>

#define MAXLEN 1000 // TODO: Work out what the real upper bound should be.
#define NODESCENT

// This requires us to start with *everything* at a given level of the census.

using namespace regina;

struct Data;
typedef std::map<std::string, Data*> Graph;
typedef Graph::iterator GraphPtr;

Graph graph;
GraphPtr TERM;
unsigned nComp;
std::queue<GraphPtr> q;
std::queue<GraphPtr> top;
// pthread_mutex_t mut;

unsigned maxTet;
unsigned long nBuild;

struct Data {
    Data* parent;
    int depth;

    Data() : parent(0), depth(0) {}
};

Data* root(Data* n) {
    Data* ans = n;
    while (ans->parent)
        ans = ans->parent;
    // Compress the tree.
    if (n == ans)
        return ans;
    Data* tmp;
    while (n->parent != ans) {
        tmp = n->parent;
        n->parent = ans;
        n = tmp;
    }
    return ans;
}

// Return value is true iff we joined two distinct components.
bool join(Data* a, Data* b) {
    Data* aRoot = root(a);
    Data* bRoot = root(b);
    if (aRoot == bRoot)
        return false;

    if (aRoot->depth > bRoot->depth)
        bRoot->parent = aRoot;
    else if (bRoot->depth > aRoot->depth)
        aRoot->parent = bRoot;
    else {
        aRoot->parent = bRoot;
        ++bRoot->depth;
    }
    return true;
}

void init() {
    // pthread_mutex_init(&mut, 0);
    TERM = graph.end();
    nComp = 0;
}

void process(const std::string sig, Data* dat) {
    Triangulation<3>* t = Triangulation<3>::fromIsoSig(sig);
    if (! t) {
        std::cerr << "BAD ISOSIG: " << sig << std::endl;
    }

    int i;
    std::string next;
    GraphPtr pos;

    if (t->size() < maxTet) {
        for (i = 0; i < t->countTriangles(); ++i) {
            Triangulation<3> copy(*t, false);
            if (copy.pachner(copy.triangle(i), true, true)) {
                next = copy.isoSig();

                // pthread_mutex_lock(&mut);
                pos = graph.find(next);
                if (pos == TERM) {
                    pos = graph.insert(Graph::value_type(next, new Data())).first;
                    q.push(pos);
                    if (copy.size() == maxTet)
                        top.push(pos);
                    if (! join(dat, pos->second))
                        std::cerr << "ERROR: 2-3 adjacency problem!" << std::endl;
                    ++nBuild;
                } else {
                    if (join(dat, pos->second))
                        --nComp;
                }
                // pthread_mutex_unlock(&mut);
            }
        }
    }

    for (i = 0; i < t->countEdges(); ++i) {
        Triangulation<3> copy(*t, false);
        if (copy.pachner(copy.edge(i), true, true)) {
            next = copy.isoSig();

            // pthread_mutex_lock(&mut);
            pos = graph.find(next);
            if (pos == TERM) {
                pos = graph.insert(Graph::value_type(next, new Data())).first;
                q.push(pos);
                if (! join(dat, pos->second))
                    std::cerr << "ERROR: 3-2 adjacency problem!" << std::endl;
                ++nBuild;
            } else {
                if (join(dat, pos->second))
                    --nComp;
            }
            // pthread_mutex_unlock(&mut);
        }
    }

    delete t;
}

void* workQueue(void*) {
    GraphPtr pos;
    std::string sig;
    Data* dat;
    while (! q.empty()) {
        {
            // pthread_mutex_lock(&mut);
            pos = q.front();
            if (pos == TERM || nComp == 1) {
                // pthread_mutex_unlock(&mut);
                break;
            }
            q.pop();
            sig = pos->first;
            dat = pos->second;
            // graph.erase(pos);
            // pthread_mutex_unlock(&mut);
        }
        process(sig, dat);
    }
    return 0;
}

int main() {
    init();

    nBuild = 0;

    char input[MAXLEN];
    scanf("%s", input);
    std::cout << "MANIFOLD: " << input << std::endl;
    while (true) {
        scanf("%s", input);
        if (feof(stdin))
            break;

        GraphPtr pos = graph.insert(Graph::value_type(input, new Data)).first;
        top.push(pos);
        ++nComp;
    }

    int height = 0;
    while (nComp > 1) {
        std::cout << nComp << " components" << std::endl;
        ++height;
        maxTet = (top.front()->first[0] - 'a') + 1;
        // q.push(TERM);
        std::swap(top, q);

        /*
        pthread_t t1; pthread_create(&t1, 0, &workQueue, 0);
        pthread_t t2; pthread_create(&t2, 0, &workQueue, 0);
        pthread_t t3; pthread_create(&t3, 0, &workQueue, 0);
        pthread_t t4; pthread_create(&t4, 0, &workQueue, 0);
        pthread_t t5; pthread_create(&t5, 0, &workQueue, 0);
        pthread_t t6; pthread_create(&t6, 0, &workQueue, 0);
        pthread_t t7; pthread_create(&t7, 0, &workQueue, 0);
        pthread_t t8; pthread_create(&t8, 0, &workQueue, 0);

        pthread_join(t1, 0);
        pthread_join(t2, 0);
        pthread_join(t3, 0);
        pthread_join(t4, 0);
        pthread_join(t5, 0);
        pthread_join(t6, 0);
        pthread_join(t7, 0);
        pthread_join(t8, 0);
        */
        workQueue(0);

        // q.pop();
    }
    std::cout << nComp << " component" << std::endl;
    std::cout << "Height, ->: " << height << ' ' << nBuild << ' '
        << std::endl;
    // pthread_mutex_destroy(&mut);

    return 0;
}

